@@ -16,12 +16,14 @@ OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
 
+#![cfg_attr(docsrs, feature(doc_cfg, doc_auto_cfg))]
+
 //! # Internationalisation helper
 //!
 //! This crate maily expose a macro that wraps gettext in a convinient ways.
-//! See the documentation of the [tr! macro](macro.tr.html).
+//! See the documentation of the [tr! macro](tr!).
 //!
-//! To translate a rust crate, simply wrap your string within the [`tr!` macro](macro.tr.html).
+//! To translate a rust crate, simply wrap your string within the [`tr!` macro](tr!).
 //! One can then use the `xtr` binary to extract all the translatable from a crate in a `.po`
 //! file. GNU gettext tools can be used to process and translate these strings.
 //!
@@ -64,9 +66,21 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //!
 //! - **`gettext-rs`** *(enabled by default)* - This crate wraps the gettext C library
 //! - **`gettext`** - A rust re-implementation of gettext. That crate does not take care of loading the
-//!   right .mo files, so one must use the (`set_translator!`)[macro.set_translator.html] macro with a
+//!   right .mo files, so one must use the [`set_translator!] macro with a
 //!   `gettext::Catalog` object
 //!
+//! Additionally, this crate permits loading from `.po` or `.mo` files directly via the [`PoTranslator`] and
+//! [`MoTranslator`] types, guarded beind the respective **`mo-translator`** and **`po-translator`** features.
+
+#[cfg(any(feature = "po-translator", feature = "mo-translator"))]
+mod rspolib_translator;
+#[cfg(any(feature = "po-translator", feature = "mo-translator"))]
+pub use rspolib_translator::MoPoTranslatorLoadError;
+
+#[cfg(feature = "mo-translator")]
+pub use rspolib_translator::MoTranslator;
+#[cfg(feature = "po-translator")]
+pub use rspolib_translator::PoTranslator;
 
 use std::borrow::Cow;
 
@@ -81,80 +95,75 @@ pub mod runtime_format {
     //!
     //! TODO: better error reporting and support for more replacement option
 
-    /// The result of the runtime_format! macro.
-    /// This implements the Display
-    pub struct FormatArg<'a> {
-        #[doc(hidden)]
-        pub format_str: &'a str,
-        #[doc(hidden)]
-        pub args: &'a [(&'static str, &'a dyn (::std::fmt::Display))],
-    }
-
-    impl<'a> ::std::fmt::Display for FormatArg<'a> {
-        fn fmt(&self, f: &mut ::std::fmt::Formatter<'_>) -> ::std::fmt::Result {
-            let mut arg_idx = 0;
-            let mut pos = 0;
-            while let Some(mut p) = self.format_str[pos..].find(|x| x == '{' || x == '}') {
-                if self.format_str.len() - pos < p + 1 {
-                    break;
-                }
-                p += pos;
-
-                // Skip escaped }
-                if self.format_str.get(p..=p) == Some("}") {
-                    self.format_str[pos..=p].fmt(f)?;
-                    if self.format_str.get(p + 1..=p + 1) == Some("}") {
-                        pos = p + 2;
-                    } else {
-                        // FIXME! this is an error, it should be reported  ('}' must be escaped)
-                        pos = p + 1;
-                    }
-                    continue;
-                }
-
-                // Skip escaped {
-                if self.format_str.get(p + 1..=p + 1) == Some("{") {
-                    self.format_str[pos..=p].fmt(f)?;
-                    pos = p + 2;
-                    continue;
-                }
-
-                // Find the argument
-                let end = if let Some(end) = self.format_str[p..].find('}') {
-                    end + p
-                } else {
-                    // FIXME! this is an error, it should be reported
-                    self.format_str[pos..=p].fmt(f)?;
-                    pos = p + 1;
-                    continue;
-                };
-                let argument = self.format_str[p + 1..end].trim();
-                let pa = if p == end - 1 {
-                    arg_idx += 1;
-                    arg_idx - 1
-                } else if let Ok(n) = argument.parse::<usize>() {
-                    n
-                } else if let Some(p) = self.args.iter().position(|x| x.0 == argument) {
-                    p
-                } else {
-                    // FIXME! this is an error, it should be reported
-                    self.format_str[pos..end].fmt(f)?;
-                    pos = end;
-                    continue;
-                };
-
-                // format the part before the '{'
-                self.format_str[pos..p].fmt(f)?;
-                if let Some(a) = self.args.get(pa) {
-                    a.1.fmt(f)?;
-                } else {
-                    // FIXME! this is an error, it should be reported
-                    self.format_str[p..=end].fmt(f)?;
-                }
-                pos = end + 1;
+    /// Converts the result of the runtime_format! macro into the final String
+    pub fn display_string(format_str: &str, args: &[(&str, &dyn ::std::fmt::Display)]) -> String {
+        use ::std::fmt::Write;
+        let fmt_len = format_str.len();
+        let mut res = String::with_capacity(2 * fmt_len);
+        let mut arg_idx = 0;
+        let mut pos = 0;
+        while let Some(mut p) = format_str[pos..].find(['{', '}']) {
+            if fmt_len - pos < p + 1 {
+                break;
             }
-            self.format_str[pos..].fmt(f)
+            p += pos;
+
+            // Skip escaped }
+            if format_str.get(p..=p) == Some("}") {
+                res.push_str(&format_str[pos..=p]);
+                if format_str.get(p + 1..=p + 1) == Some("}") {
+                    pos = p + 2;
+                } else {
+                    // FIXME! this is an error, it should be reported  ('}' must be escaped)
+                    pos = p + 1;
+                }
+                continue;
+            }
+
+            // Skip escaped {
+            if format_str.get(p + 1..=p + 1) == Some("{") {
+                res.push_str(&format_str[pos..=p]);
+                pos = p + 2;
+                continue;
+            }
+
+            // Find the argument
+            let end = if let Some(end) = format_str[p..].find('}') {
+                end + p
+            } else {
+                // FIXME! this is an error, it should be reported
+                res.push_str(&format_str[pos..=p]);
+                pos = p + 1;
+                continue;
+            };
+            let argument = format_str[p + 1..end].trim();
+            let pa = if p == end - 1 {
+                arg_idx += 1;
+                arg_idx - 1
+            } else if let Ok(n) = argument.parse::<usize>() {
+                n
+            } else if let Some(p) = args.iter().position(|x| x.0 == argument) {
+                p
+            } else {
+                // FIXME! this is an error, it should be reported
+                res.push_str(&format_str[pos..end]);
+                pos = end;
+                continue;
+            };
+
+            // format the part before the '{'
+            res.push_str(&format_str[pos..p]);
+            if let Some(a) = args.get(pa) {
+                write!(&mut res, "{}", a.1)
+                    .expect("a Display implementation returned an error unexpectedly");
+            } else {
+                // FIXME! this is an error, it should be reported
+                res.push_str(&format_str[p..=end]);
+            }
+            pos = end + 1;
         }
+        res.push_str(&format_str[pos..]);
+        res
     }
 
     #[doc(hidden)]
@@ -163,16 +172,13 @@ pub mod runtime_format {
     macro_rules! runtime_format {
         ($fmt:expr) => {{
             // TODO! check if 'fmt' does not have {}
-            format!("{}", $fmt)
+            String::from($fmt)
         }};
         ($fmt:expr,  $($tail:tt)* ) => {{
-            let format_str = $fmt;
-            let fa = $crate::runtime_format::FormatArg {
-                format_str: AsRef::as_ref(&format_str),
-                //args: &[ $( $crate::runtime_format!(@parse_arg $e) ),* ],
-                args: $crate::runtime_format!(@parse_args [] $($tail)*)
-            };
-            format!("{}", fa)
+            $crate::runtime_format::display_string(
+                AsRef::as_ref(&$fmt),
+                $crate::runtime_format!(@parse_args [] $($tail)*),
+            )
         }};
 
         (@parse_args [$($args:tt)*]) => { &[ $( $args ),* ]  };
@@ -241,7 +247,7 @@ pub mod runtime_format {
 /// The backend is only responsable to provide a matching string, the formatting is done
 /// using this string.
 ///
-/// The translator for a crate can be set with the set_translator! macro
+/// The translator for a crate can be set with the [`set_translator!`] macro
 pub trait Translator: Send + Sync {
     fn translate<'a>(&'a self, string: &'a str, context: Option<&'a str>) -> Cow<'a, str>;
     fn ntranslate<'a>(
@@ -253,17 +259,34 @@ pub trait Translator: Send + Sync {
     ) -> Cow<'a, str>;
 }
 
+impl<T: Translator> Translator for std::sync::Arc<T> {
+    fn translate<'a>(
+        &'a self,
+        string: &'a str,
+        context: Option<&'a str>,
+    ) -> std::borrow::Cow<'a, str> {
+        <T as Translator>::translate(self, string, context)
+    }
+
+    fn ntranslate<'a>(
+        &'a self,
+        n: u64,
+        singular: &'a str,
+        plural: &'a str,
+        context: Option<&'a str>,
+    ) -> std::borrow::Cow<'a, str> {
+        <T as Translator>::ntranslate(self, n, singular, plural, context)
+    }
+}
+
 #[doc(hidden)]
 pub mod internal {
 
     use super::Translator;
-    use std::{borrow::Cow, collections::HashMap, sync::RwLock};
+    use std::{borrow::Cow, collections::HashMap, sync::LazyLock, sync::RwLock};
 
-    // TODO: use parking_lot::RwLock
-    lazy_static::lazy_static! {
-        static ref TRANSLATORS: RwLock<HashMap<&'static str, Box<dyn Translator>>> =
-            Default::default();
-    }
+    static TRANSLATORS: LazyLock<RwLock<HashMap<&'static str, Box<dyn Translator>>>> =
+        LazyLock::new(Default::default);
 
     pub fn with_translator<T>(module: &'static str, func: impl FnOnce(&dyn Translator) -> T) -> T {
         let domain = domain_from_module(module);
@@ -288,7 +311,7 @@ pub mod internal {
     }
     #[cfg(feature = "gettext-rs")]
     fn demangle_context(r: String) -> String {
-        if let Some(x) = r.split('\u{4}').last() {
+        if let Some(x) = r.split('\u{4}').next_back() {
             return x.to_owned();
         }
         r
@@ -300,7 +323,7 @@ pub mod internal {
     impl Translator for DefaultTranslator {
         fn translate<'a>(&'a self, string: &'a str, context: Option<&'a str>) -> Cow<'a, str> {
             Cow::Owned(if let Some(ctx) = context {
-                demangle_context(gettextrs::dgettext(self.0, &mangle_context(ctx, string)))
+                demangle_context(gettextrs::dgettext(self.0, mangle_context(ctx, string)))
             } else {
                 gettextrs::dgettext(self.0, string)
             })
@@ -317,8 +340,8 @@ pub mod internal {
             Cow::Owned(if let Some(ctx) = context {
                 demangle_context(gettextrs::dngettext(
                     self.0,
-                    &mangle_context(ctx, singular),
-                    &mangle_context(ctx, plural),
+                    mangle_context(ctx, singular),
+                    mangle_context(ctx, plural),
                     n,
                 ))
             } else {
@@ -358,10 +381,16 @@ pub mod internal {
     }
 
     pub fn set_translator(module: &'static str, translator: impl Translator + 'static) {
+        let domain = domain_from_module(module);
         TRANSLATORS
             .write()
             .unwrap()
-            .insert(module, Box::new(translator));
+            .insert(domain, Box::new(translator));
+    }
+
+    pub fn unset_translator(module: &'static str) {
+        let domain = domain_from_module(module);
+        TRANSLATORS.write().unwrap().remove(domain);
     }
 }
 
@@ -480,7 +509,7 @@ macro_rules! tr_init {
 
 /// Set the translator to be used for this crate.
 ///
-/// The argument needs to be something implementing the Translator trait
+/// The argument needs to be something implementing the [`Translator`] trait
 ///
 /// For example, using the gettext crate (if the gettext feature is enabled)
 /// ```ignore
@@ -495,6 +524,16 @@ macro_rules! set_translator {
     };
 }
 
+/// Clears the translator to be used for this crate.
+///
+/// Use this macro to return back to the source language.
+#[macro_export]
+macro_rules! unset_translator {
+    () => {
+        $crate::internal::unset_translator(module_path!())
+    };
+}
+
 #[cfg(feature = "gettext")]
 impl Translator for gettext::Catalog {
     fn translate<'a>(&'a self, string: &'a str, context: Option<&'a str>) -> Cow<'a, str> {
@@ -544,5 +583,10 @@ mod tests {
             tr!("ctx" => "{0} have one item" | "{0} have {n} items" % 42, "I"),
             "I have 42 items"
         );
+
+        assert_eq!(
+            tr!("{} = {}", 255, format_args!("{:#x}", 255)),
+            "255 = 0xff"
+        );
     }
 }
